(*
  Anonymous method completion plugin

  Copyright (c) 2015 Lyna

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)
unit UAnonComplete;

interface

uses
  System.SysUtils, System.Classes, Vcl.Menus, ToolsAPI, RegularExpressions;

type
  TKeyboardBinding = class(TNotifierObject, IOTAKeyboardBinding)
  private
    function GetBindingType: TBindingType;
    function GetDisplayName: string;
    function GetName: string;
    procedure BindKeyboard(const BindingServices: IOTAKeyBindingServices);

    procedure AnonMethodComplete(const Context: IOTAKeyContext; KeyCode: TShortcut;
      var BindingResult: TKeyBindingResult);
  end;

const
  DefinedTypes: array[0..80] of record Name, Prototype: string end = (
    (Name: 'TStringTransform'; Prototype: 'procedure(const InStr: string; var OutStr: string)'),
    (Name: 'TDBXScheduleEvent'; Prototype: 'procedure'),
    (Name: 'TObjectsConverter'; Prototype: 'function(Data: TObject; Field: string): TListOfObjects'),
    (Name: 'TStringsConverter'; Prototype: 'function(Data: TObject; Field: string): TListOfStrings'),
    (Name: 'TTypeObjectsConverter'; Prototype: 'function(Data: TObject): TListOfObjects'),
    (Name: 'TTypeStringsConverter'; Prototype: 'function(Data: TObject): TListOfStrings'),
    (Name: 'TObjectConverter'; Prototype: 'function(Data: TObject; Field: string): TObject'),
    (Name: 'TStringConverter'; Prototype: 'function(Data: TObject; Field: string): string'),
    (Name: 'TTypeObjectConverter'; Prototype: 'function(Data: TObject): TObject'),
    (Name: 'TTypeStringConverter'; Prototype: 'function(Data: TObject): string'),
    (Name: 'TObjectsReverter'; Prototype: 'procedure(Data: TObject; Field: string; Args: TListOfObjects)'),
    (Name: 'TStringsReverter'; Prototype: 'procedure(Data: TObject; Field: string; Args: TListOfStrings)'),
    (Name: 'TTypeObjectsReverter'; Prototype: 'function(Data: TListOfObjects): TObject'),
    (Name: 'TTypeStringsReverter'; Prototype: 'function(Data: TListOfStrings): TObject'),
    (Name: 'TObjectReverter'; Prototype: 'procedure(Data: TObject; Field: string; Arg: TObject)'),
    (Name: 'TStringReverter'; Prototype: 'procedure(Data: TObject; Field: string; Arg: string)'),
    (Name: 'TTypeObjectReverter'; Prototype: 'function(Data: TObject): TObject'),
    (Name: 'TTypeStringReverter'; Prototype: 'function(Data: string): TObject'),
    (Name: 'TFormatCallback'; Prototype: 'procedure(const ASourceScope, AControlScope: IScope)'),
    (Name: 'TQueryBreakCallback'; Prototype: 'function(const ASourceScope: IScope): Boolean'),
    (Name: 'TBindGridOnCallbackEvaluateFormat'; Prototype: 'procedure(Col, Row: Integer; AEditorScope: IScope)'),
    (Name: 'TBindCompDesignExpression.TExecuteDesignExpression'; Prototype: 'procedure(const AName: string; const AExpression: string; ACallback: TProc<IValue>)'),
    (Name: 'TBindCompDesignExpression.TExecuteDesignExpression2'; Prototype: 'procedure(const AName: string; const AControlExpression, ASourceExpression: string; ACallback: TProc<IValue>)'),
    (Name: 'TBindCompDesignExpression.TSaveDesignExpression'; Prototype: 'procedure(const AName: string; const AExpression: string)'),
    (Name: 'TValueGeneratorOnGetValue'; Prototype: 'function(AKey: TValue; var AFreeObject: Boolean): TValue'),
    (Name: 'TCustomScope.TWrapperFoundCallback'; Prototype: 'procedure(Scope: TCustomScope; const Wrapper: ICustomWrapper)'),
    (Name: 'TSubscriptionNotification'; Prototype: 'procedure(const ASubscription: ISubscription)'),
    (Name: 'TCustomWrapperGetterCallback'; Prototype: 'function(ParentObject: TObject; const MemberName: string; Args: TArray<TValue>): TValue'),
    (Name: 'TLocationWrapper.TValueGetter'; Prototype: 'function: TValue'),
    (Name: 'TLocationWrapper.TValueSetter'; Prototype: 'procedure(AValue: TValue)'),
    (Name: 'TValueInvokableBody'; Prototype: 'function(Args: TArray<TValue>): TValue'),
    (Name: 'TInvokableBody'; Prototype: 'function(Args: TArray<IValue>): IValue'),
    (Name: 'TBindingCallbackFactory.TBindingCallback'; Prototype: 'function(const Obj: TObject; const Val: TValue): Boolean'),
    (Name: 'TConvertProc'; Prototype: 'procedure(const InValue: TValue; var OutValue: TValue)'),
    (Name: 'TConvertProc2'; Prototype: 'procedure(const InValue: TValue; ATo: PTypeInfo; var OutValue: TValue)'),
    (Name: 'TGraph.TFindGraphNodeCallback'; Prototype: 'function(Node: TGraphNode): Boolean'),
    (Name: 'TEnumControlsRef'; Prototype: 'procedure(const AControl: TControl; var Done: Boolean)'),
    (Name: 'TPropertyApplyProc'; Prototype: 'procedure(Instance: TObject; Prop: TRttiProperty)'),
    (Name: 'TInputCloseQueryFunc'; Prototype: 'function(const Values: array of string): Boolean'),
    (Name: 'TMessageListener'; Prototype: 'procedure(const Sender: TObject; const M: TMessage)'),
    (Name: 'TFmxObjectSortCompare'; Prototype: 'function(Left, Right: TFmxObject): Integer'),
    (Name: 'Log.ToStringFunc'; Prototype: 'function(O: TObject): string'),
    (Name: 'TExtensionMethod'; Prototype: 'function(AExt: PChar; ALength: Integer): Boolean'),
    (Name: 'TEnumActionListRef'; Prototype: 'procedure(const Action: TContainedAction; var Done: Boolean)'),
    (Name: 'TListSortCompareFunc'; Prototype: 'function(Item1, Item2: Pointer): Integer'),
    (Name: 'TThreadProcedure'; Prototype: 'procedure'),
    (Name: 'TObserverToggleEvent'; Prototype: 'procedure(const AObserver: IObserver; const Value: Boolean)'),
    (Name: 'TObserverGetCurrentEvent'; Prototype: 'function: TVarRec'),
    (Name: 'TObservers.TCanObserveEvent'; Prototype: 'function(const ID: Integer): Boolean'),
    (Name: 'TObservers.TObserverAddedEvent'; Prototype: 'procedure(const ID: Integer; const Observer: IObserver)'),
    (Name: 'TLoginCredentialService.TLoginFunc'; Prototype: 'function(const Username, Password, Domain: string): Boolean'),
    (Name: 'TPostFoundCycleProc'; Prototype: 'procedure(const ClassName: string; Reference: IntPtr; const Stack: TStack<IntPtr>)'),
    (Name: 'TList<T>.TEmptyFunc'; Prototype: 'function(const L, R: \1): Boolean'),
    (Name: 'TList<T>.TListCompareFunc'; Prototype: 'function(const L, R: \1): Integer'),
    (Name: 'TComparison<T>'; Prototype: 'function(const Left, Right: \1): Integer'),
    (Name: 'TEqualityComparison<T>'; Prototype: 'function(const Left, Right: \1): Boolean'),
    (Name: 'THasher<T>'; Prototype: 'function(const Value: \1): Integer'),
    (Name: 'TDirectory.TFilterPredicate'; Prototype: 'function(const Path: string; const SearchRec: TSearchRec): Boolean'),
    (Name: 'TDirectory.TDirectoryWalkProc'; Prototype: 'function(const Path: string; const FileInfo: TSearchRec): Boolean'),
    (Name: 'TMethodImplementationCallback'; Prototype: 'procedure(UserData: Pointer; const Args: TArray<TValue>; out Result: TValue)'),
    (Name: 'TInterceptBeforeNotify'; Prototype: 'procedure(Instance: TObject; Method: TRttiMethod; const Args: TArray<TValue>; out DoInvoke: Boolean; out Result: TValue)'),
    (Name: 'TInterceptAfterNotify'; Prototype: 'procedure(Instance: TObject; Method: TRttiMethod; const Args: TArray<TValue>; var Result: TValue)'),
    (Name: 'TInterceptExceptionNotify'; Prototype: 'procedure(Instance: TObject; Method: TRttiMethod; const Args: TArray<TValue>; out RaiseException: Boolean; TheException: Exception; out Result: TValue)'),
    (Name: 'TVirtualInterfaceInvokeEvent'; Prototype: 'procedure(Method: TRttiMethod; const Args: TArray<TValue>; out Result: TValue)'),
    (Name: 'TSensorFilter'; Prototype: 'function(Sensor: TCustomSensor): Boolean'),
    (Name: 'TEventDispatchInvoker'; Prototype: 'procedure(DispId: Integer; var Params: TOleVariantArray)'),
    (Name: 'TStreamConstructor'; Prototype: 'function(InStream: TStream; const ZipFile: TZipFile; const Item: TZipHeader): TStream'),
    (Name: 'TProc'; Prototype: 'procedure'),
    (Name: 'TProc<T>'; Prototype: 'procedure(Arg1: \1)'),
    (Name: 'TProc<T,T>'; Prototype: 'procedure(Arg1: \1; Arg2: \2)'),
    (Name: 'TProc<T,T,T>'; Prototype: 'procedure(Arg1: \1; Arg2: \2; Arg3: \3)'),
    (Name: 'TProc<T,T,T,T>'; Prototype: 'procedure(Arg1: \1; Arg2: \2; Arg3: \3; Arg4: \4)'),
    (Name: 'TFunc<T>'; Prototype: 'function: \1'),
    (Name: 'TFunc<T,T>'; Prototype: 'function(Arg1: \1): \2'),
    (Name: 'TFunc<T,T,T>'; Prototype: 'function(Arg1: \1; Arg2: \2): \3'),
    (Name: 'TFunc<T,T,T,T>'; Prototype: 'function(Arg1: \1; Arg2: \2; Arg3: \3): \4'),
    (Name: 'TFunc<T,T,T,T,T>'; Prototype: 'function(Arg1: \1; Arg2: \2; Arg3: \3; Arg4: \4): \5'),
    (Name: 'TPredicate<T>'; Prototype: 'function(Arg1: \1): Boolean'),
    (Name: 'TCheckProc'; Prototype: 'function(const Directory: string): Boolean'),
    (Name: 'TInputCloseQueryFunc'; Prototype: 'function(const Values: array of string): Boolean'),
    (Name: 'TGestureProviderChangeEvent'; Prototype: 'procedure(Sender: TObject; AGesture: TCustomGestureCollectionItem)')
  );

implementation

{ TKeyboardBinding }

procedure TKeyboardBinding.AnonMethodComplete(const Context: IOTAKeyContext;
  KeyCode: TShortcut; var BindingResult: TKeyBindingResult);
var
  codeInsight: IOTACodeInsightServices;
  codeInsightMan: IOTACodeInsightManager;
  paramList: IOTACodeInsightParameterList;
  i, paramIndex: Integer;
  query: IOTACodeInsightParamQuery;
  arr: TArray<string>;
  m: TMatch;
  value: string;

  function GetValueByName(const Name: string; out Value: string): Boolean;
  var
    i, j: Integer;
    m: TMatch;
    s: string;
    p, p1: PChar;
    nest: Integer;
    typeParams: TArray<string>;
    dummy: TArray<string>;

    function RemoveUnitNames(const Value: string): string;
    begin
      Result := TRegEx.Replace(Value, '[^,<>]+\.', '');
    end;

  begin
    m := TRegEx.Match(Name, '(.*?)<(.*)>(.*)');
    if m.Success then
    begin
      s := m.Groups[2].Value;
      p := PChar(s);
      p1 := p;
      nest := 0;
      while p1^ <> #0 do
      begin
        case p1^ of
          '<': Inc(nest);
          '>': Dec(nest);
          ',':
          begin
            if nest = 0 then
            begin
              SetLength(typeParams, Length(typeParams) + 1);
              SetString(typeParams[High(typeParams)], p, p1 - p);
              p := p1 + 1;
            end;
          end;
        end;
        Inc(p1);
      end;
      if p <> p1 then
      begin
        SetLength(typeParams, Length(typeParams) + 1);
        SetString(typeParams[High(typeParams)], p, p1 - p);
      end;

      SetLength(dummy, Length(typeParams));
      for i := Low(dummy) to High(dummy) do
        dummy[i] := 'T';
      s := m.Groups[1].Value + '<' + string.Join(',', dummy) + '>' + m.Groups[3].Value;
    end
    else begin
      s := Name;
    end;

    for i := Low(DefinedTypes) to High(DefinedTypes) do
    begin
      if not s.EndsWith(DefinedTypes[i].Name, True) then Continue;

      Value := DefinedTypes[i].Prototype;
      for j := Low(typeParams) to High(typeParams) do
        Value := Value.Replace('\'+(j+1).ToString, RemoveUnitNames(typeParams[j]));

      Exit(True);
    end;

    Result := False;
  end;

begin
  BindingResult := krUnhandled;

  codeInsight := BorlandIDEServices as IOTACodeInsightServices;
  codeInsight.GetCurrentCodeInsightManager(codeInsightMan);
  if codeInsightMan = nil then Exit;

  if not codeInsightMan.Enabled then Exit;
  try
    if not codeInsightMan.InvokeParameterCodeInsight(itManual, paramIndex) then Exit;
  except
    Exit;
  end;

  codeInsightMan.GetParameterList(paramList);
  if paramList = nil then Exit;
  for i := 0 to paramList.ProcedureCount-1 do
  begin
    paramList.GetParameterQuery(i, query);
    if query = nil then Continue;

    arr := paramList.ProcedureParamsText[i].Split([#13#10], TStringSplitOptions.ExcludeEmpty);
    if paramIndex >= Length(arr) then Continue;

    m := TRegEx.Match(arr[paramIndex], '.*?\s?.*?: (.*)');
    if m.Success and GetValueByName(m.Groups[1].Value, value) then
    begin
      Context.EditBuffer.EditPosition.InsertText(#13#10);
      Context.EditBuffer.EditPosition.InsertCharacter(#9);
      Context.EditBuffer.EditPosition.InsertText(value+#13#10'begin'#13#10'end');
      BindingResult := krHandled;
      Exit;
    end;
  end;
end;

procedure TKeyboardBinding.BindKeyboard(
  const BindingServices: IOTAKeyBindingServices);
begin
  BindingServices.AddKeyBinding([TextToShortCut('Ctrl+Alt+Space')], AnonMethodComplete, nil);
end;

function TKeyboardBinding.GetBindingType: TBindingType;
begin
  Result := btPartial;
end;

function TKeyboardBinding.GetDisplayName: string;
begin
  Result := 'Anonymous method completion';
end;

function TKeyboardBinding.GetName: string;
begin
  Result := 'AnonymousMethodCompletion';
end;

var
  KeyBindingIndex: Integer;
initialization
  KeyBindingIndex := (BorlandIDEServices as IOTAKeyboardServices).AddKeyboardBinding(TKeyboardBinding.Create);
finalization
  (BorlandIDEServices as IOTAKeyboardServices).RemoveKeyboardBinding(KeyBindingIndex);
end.
